// naumann@stce.rwth-aachen.de

#pragma once

#include "config.h"
#include "objective.h"

#include "dco.hpp"

// calibration method determined at link time
template<typename T>
VT<T> fit(VT<T> p, VT<T> xobs, MT<T> yobs);

template<typename T>
MT<T> dp_dobs_sym(VT<T> p, VT<T> xobs, MT<T> yobs) {
  assert(xobs.size()==nobs);
  assert(yobs.rows()==nobs);
  assert(yobs.cols()==n);
  assert(p.size()==np);
  return dde_dp_dp(p,xobs,yobs).ldlt().solve(-dde_dp_dobs(p,xobs,yobs));
}

template<typename T>
MT<T> dp_dobs_cfd(VT<T> p, VT<T> xobs, MT<T> yobs) {
  assert(xobs.size()==nobs);
  assert(yobs.rows()==nobs);
  assert(yobs.cols()==n);
  assert(p.size()==np);
  MT<T> dpdobs(np,(n+1)*nobs);
  int i=0, k=0;
  while (i<nobs) {
    T xobs_i_save=xobs(i); 
    VT<T> pp(np), pm(np); 
    T h=eps;
    if (xobs(i)!=0) {
      h=h*fabs(xobs(i));
    }
    xobs(i)=xobs(i)+h;
    pp=fit(p,xobs,yobs);
    xobs(i)=xobs_i_save-h;
    pm=fit(p,xobs,yobs);
    xobs(i)=xobs_i_save;
    dpdobs.col(k)=(pp-pm)/(2*h);
    k=k+1;
    int j=0;
    while (j<n) {
      T yobs_ij_save=yobs(i,j); 
      h=eps;
      if (yobs(i,j)!=0) {
        h=h*fabs(yobs(i,j));
      }
      yobs(i,j)=yobs(i,j)+h;
      pp=fit(p,xobs,yobs);
      yobs(i,j)=yobs_ij_save-h;
      pm=fit(p,xobs,yobs);
      yobs(i,j)=yobs_ij_save;
      dpdobs.col(k)=(pp-pm)/(2*h);
      k=k+1;
      j=j+1;
    }
    i=i+1;
  }
  return dpdobs;
}

template<typename T>
MT<T> dp_dobs_tad(VT<T> p, VT<T> xobs, MT<T> yobs) {
  assert(xobs.size()==nobs);
  assert(yobs.rows()==nobs);
  assert(yobs.cols()==n);
  assert(p.size()==np);
  MT<T> dpdobs(np,(n+1)*nobs);
  using TT=dco::gt1s<T>::type;
  VT<TT> p_TT(np); 
  VT<TT> xobs_TT(nobs); 
  MT<TT> yobs_TT(nobs,n);
  int i,j,k,l;
  i=0;
  while (i<np) {
    p_TT(i)=p(i);
    i=i+1;
  }
  i=0;
  while (i<nobs) {
    xobs_TT(i)=xobs(i); 
    j=0;
    while (j<n) {
      yobs_TT(i,j)=yobs(i,j);
      j=j+1;
    }
    i=i+1;
  }
  i=0;
  k=0;
  while (i<nobs) {
    dco::derivative(xobs_TT(i))=1;
    p_TT=fit(p_TT,xobs_TT,yobs_TT);
    l=0;
    while (l<np) {
      dpdobs(l,k)=dco::derivative(p_TT(l));
      l=l+1;
    }
    k=k+1;
    dco::derivative(xobs_TT(i))=0;
    l=0;
    while (l<np) {
      p_TT(l)=p(l);
      l=l+1;
    }
    j=0;
    while (j<n) {
      dco::derivative(yobs_TT(i,j))=1;
      p_TT=fit(p_TT,xobs_TT,yobs_TT);
      l=0;
      while (l<np) {
        dpdobs(l,k)=dco::derivative(p_TT(l));
        l=l+1;
      }
      dco::derivative(yobs_TT(i,j))=0;
      l=0;
      while (l<np) {
        p_TT(l)=p(l);
        l=l+1;
      }
      k=k+1;
      j=j+1;
    }
    i=i+1;
  }
  return dpdobs;
}

template<typename T>
MT<T> dp_dobs_aad(VT<T> p, VT<T> xobs, MT<T> yobs) {
  // conditions
  assert(xobs.size()==nobs);
  assert(yobs.rows()==nobs);
  assert(yobs.cols()==n);
  assert(p.size()==np);
  using AT=dco::ga1s<T>::type;
  dco::smart_tape_ptr_t<dco::mode<AT>> tape;
  VT<AT> p_AT(np); 
  VT<AT> xobs_AT(nobs); 
  MT<AT> yobs_AT(nobs,n);
  // activation and registration
  int i=0;
  while (i<np) {
    p_AT(i)=p(i);
    i=i+1;
  }
  i=0;
  while (i<nobs) {
    xobs_AT(i)=xobs(i); 
    tape->register_variable(xobs_AT(i));
    int j=0;
    while (j<n) {
      yobs_AT(i,j)=yobs(i,j);
      tape->register_variable(yobs_AT(i,j));
      j=j+1;
    }
    i=i+1;
  }
  // recording
  p_AT=fit(p_AT,xobs_AT,yobs_AT);
  // to be returned
  MT<T> dpdobs(np,(n+1)*nobs);
  int k=0;
  while (k<np) {
    // seeding for k-th row of Jacobian
    dco::derivative(p_AT(k))=1;
    // interpretation of tape
    tape->interpret_adjoint();
    // harvesting of k-th row of Jacobian
    int l=0;
    i=0;
    while (i<nobs) {
      dpdobs(k,l)=dco::derivative(xobs_AT(i));
      l=l+1;
      int j=0;
      while (j<n) {
        dpdobs(k,l)=dco::derivative(yobs_AT(i,j));
        l=l+1;
        j=j+1; 
      }
      i=i+1;
    }
    k=k+1;
    // preparation for reuse of vector of adjoints
    tape->zero_adjoints(); 
  }
  return dpdobs;
}

