// naumann@stce.rwth-aachen.de

#pragma once

#include "config.h"
#include "objective.h"

/// stationary point of objective wrt. model parameteris by the gradient descent method
template<typename T>
VT<T> fit(VT<T> p, VT<T> xobs, MT<T> yobs) {
  // conditions
  assert(p.size()==np);
  assert(xobs.size()==nobs);
  assert(yobs.rows()==nobs);
  assert(yobs.cols()==n);
  // derivative (gradient) of objective wrt. model parameters
  VT<T> dedp=de_dp(p,xobs,yobs);
  int i=0;
  do {
    // simple line search
    T alpha=1;
    T e_=e(p,xobs,yobs);
    VT<T> p_trial=p-alpha*dedp;
    while ((e(p_trial,xobs,yobs)>=e_)&&(alpha>eps)) { // simple line search
      alpha=alpha/2; 
      p_trial=p-alpha*dedp;
    }
    // result of last gradient descent step
    p=p_trial;
    dedp=de_dp(p,xobs,yobs);
    i=i+1;
  } while (dedp.norm()>eps&&i<pow(10,6)); // (guaranteed) convergence criterion
  return p;
}
