// naumann@stce.rwth-aachen.de

#pragma once

#include "config.h"
#include "objective.h"

template<typename T>
VT<T> prefit(VT<T> p, VT<T> xobs, MT<T> yobs) {
  assert(p.size()==np);
  assert(xobs.size()==nobs);
  assert(yobs.rows()==nobs);
  assert(yobs.cols()==n);
  VT<T> dedp=de_dp(p,xobs,yobs);
  do {
    T alpha=1;
    T e_=e(p,xobs,yobs);
    VT<T> p_trial=p-alpha*dedp;
    while ((e(p_trial,xobs,yobs)>=e_)&&(alpha>eps)) {
      alpha=alpha/2; // line search
      p_trial=p-alpha*dedp;
    }
    p=p_trial;
    dedp=de_dp(p,xobs,yobs);
  } while (dedp.norm()>1e-2);
  return p;
}

template<typename T>
VT<T> fit(VT<T> p, VT<T> xobs, MT<T> yobs) {
  assert(p.size()==np);
  assert(xobs.size()==nobs);
  assert(yobs.rows()==nobs);
  assert(yobs.cols()==n);
  p=prefit(p,xobs,yobs);
  VT<T> dedp=de_dp(p,xobs,yobs);
  int i=0;
  do {
    MT<T> ddedpp=dde_dp_dp(p,xobs,yobs);
    p=p+ddedpp.ldlt().solve(-dedp);
    dedp=de_dp(p,xobs,yobs);
    i=i+1;
  } while ((dedp.norm()>eps)&&(i<1000));
  return p;
}
