// naumann@stce.rwth-aachen.de

#pragma once

#include "config.h"

/// linear regression by Householder reflection
template<typename T>
VT<T> linear_regression(MT<T> drdp, VT<T> yobs) {
  // conditions
  assert(yobs.size()==nobs*n); assert(drdp.rows()==nobs*n); assert(drdp.cols()==np);
  // factorization
  MT<T> R=drdp, Q_transpose=MT<T>::Identity(nobs*n,nobs*n);
  VT<T> p(np);
  int i,j;
  j=0;
  while (j<np) {
    VT<T> x=R.col(j);
    // ensures identity as leading principal (j x j)-submatrix of H 
    i=0;
    while (i<j) {
      x(i)=0;
      i=i+1;
    }
    // computation of j-th Householder matrix
    VT<T> v=x;
    v(j)=v(j)-x(j)/fabs(x(j))*x.norm();
    VT<T> c=2*v/(v.dot(v));
    MT<T> H=Eigen::MatrixX<T>::Identity(nobs*n,nobs*n)-2*v*v.transpose()/v.dot(v);
    // update of Q^T
    Q_transpose=Q_transpose-c*(v.transpose()*Q_transpose);
    // update of R
    R=R-c*(v.transpose()*R);
    j=j+1;
  }
  // solution (backward substitution)
  VT<T> r=Q_transpose*yobs;
  i=np-1;
  while (i>=0) {
    j=np-1;
    while (j>i) {
      r(i)=r(i)-R(i,j)*p(j);
      j=j-1;
    }
    p(i)=r(i)/R(i,i);
    i=i-1;
  }
  return p;
}

