// naumann@stce.rwth-aachen.de

#pragma once

#include "config.h"
#include "objective.h"

/// stationary point of objective wrt. model parameter by Newton method
template<typename T>
VT<T> fit(VT<T> p, VT<T> xobs, MT<T> yobs) {
  // conditions
  assert(p.size()==np); assert(xobs.size()==nobs); 
  assert(yobs.rows()==nobs); assert(yobs.cols()==n);
  // gradient of objective wrt. model parameter
  VT<T> dedp=de_dp(p,xobs,yobs);
  int i=0;
  do {
    // Newton step requires Hessian of objective wrt. model parameter
    MT<T> ddedpp=dde_dp_dp(p,xobs,yobs);
    // Newton step solved by LDLT factorization to account for
    // potential indefiniteness of the Hessian 
    p=p+ddedpp.ldlt().solve(-dedp);
    dedp=de_dp(p,xobs,yobs);
    i=i+1;
  } while ((dedp.norm()>eps)&&(i<1000)); // guaranteed to terminate
  return p;
}
