// naumann@stce.rwth-aachen.de

#pragma once

#include "config.h"
#include "residual.h"
#include "objective.h"

// linear regression method (normal equation or Householder)
template<typename T>
VT<T> linear_regression(MT<T> drdp, VT<T> yobs);

// stationary point of objective wrt. model parameter by nonlinear regression
template<typename T>
VT<T> fit(VT<T> p, VT<T> xobs, MT<T> yobs) {
  // conditions
  assert(p.size()==np);
  assert(xobs.size()==nobs);
  assert(yobs.rows()==nobs);
  assert(yobs.cols()==n);
  T e_=e(p,xobs,yobs);
  do {
    VT<T> r_=r(p,xobs,yobs);
    MT<T> drdp=dr_dp(p,xobs,yobs);
    // linearization
    VT<T> delta_p=linear_regression(drdp,r_);
    // simple line search
    T e_prev=e_, alpha=2;
    do {
      VT<T> p_trial=p;
      alpha=alpha/2;
      p_trial=p_trial-alpha*delta_p;
      e_=e(p,xobs,yobs);
    } while (e_prev<e_);
    // iterative update
    p=p-alpha*delta_p;
  } while (de_dp(p,xobs,yobs).norm()>eps);
  return p;
}
