// naumann@stce.rwth-aachen.de

#include "config.h"
#include "objective.h"
#include "error_analysis.h"

#include <iostream>
using namespace std;

template<typename T>
VT<T> fit(VT<T> p, VT<T> xobs, MT<T> yobs);

int main() {
  using T=double;
  VT<T> p=VT<T>::Ones(np); // initial guess
  VT<T> xobs=VT<T>::Random(nobs); // random entries in [-1,1]
  xobs=xobs.cwiseProduct(xobs); // random entries in [0,1]
  MT<T> yobs=MT<T>::Random(nobs,n); // random entries in [-1,1]
  p=fit(p,xobs,yobs);
  cout << "p=" << p.transpose() << endl;
  cout << "||dedp||=" << de_dp(p,xobs,yobs).norm() << endl;
  cout << "spd(ddedpp)=" << (dde_dp_dp(p,xobs,yobs).llt().info()!=Eigen::NumericalIssue) << endl;
  cout << "dp_dobs (symbolic)=" << endl << dp_dobs_sym(p,xobs,yobs).norm() << endl;
/*
  bool s; 
  cout << "stability analysis by central finite differences? (yes: 1 | no: 0) "; 
  cin >> s;
  if (s) {
    p=VT<T>::Ones(np); 
    cout << "dp_dobs (cfd)=" << endl << dp_dobs_cfd(p,xobs,yobs).norm() << endl;
  }
  cout << "stability analysis by tangent AD? (yes: 1 | no: 0) "; 
  cin >> s;
  if (s) {
    p=VT<T>::Ones(np); 
    cout << "dp_dobs (tangent AD)=" << endl << dp_dobs_tad(p,xobs,yobs).norm() << endl;
  }
  cout << "stability analysis by adjoint AD? (yes: 1 | no: 0) "; 
  cin >> s;
  if (s) {
    p=VT<T>::Ones(np); 
    cout << "dp_dobs (adjoint AD)=" << endl << dp_dobs_aad(p,xobs,yobs).norm() << endl;
  }
*/
  return 0;
}

